"""

Vulture sometimes reports used code as unused. To avoid these
false-positives, you can write a Python file that explicitly uses the
code and pass it to vulture:

vulture myscript.py mydir mywhitelist.py

When creating a whitelist file, you have to make sure not to write code
that hides unused code in other files. E.g., this is why we don't import
and access the "sys" module below. If we did import it, vulture would
not be able to detect whether other files import "sys" without using it.

This file explicitly uses code from the Python standard library that is
often incorrectly detected as unused.

"""


class Whitelist:
    """
    Helper class that allows mocking Python objects.

    Use it to create whitelist files that are not only syntactically
    correct, but can also be executed.

    """

    def __getattr__(self, _):
        pass


assert Whitelist



"""
This module handles retrieval of configuration values from either the
command-line arguments or the pyproject.toml file.
"""

import argparse
import pathlib

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from .version import __version__

#: Possible configuration options and their respective defaults
DEFAULTS = {
    "config": "pyproject.toml",
    "min_confidence": 0,
    "paths": [],
    "exclude": [],
    "ignore_decorators": [],
    "ignore_names": [],
    "make_whitelist": False,
    "sort_by_size": False,
    "verbose": False,
}


class InputError(Exception):
    def __init__(self, message):
        self.message = message


def _check_input_config(data):
    """
    Checks the types of the values in *data* against the expected types of
    config-values. If a value has the wrong type, raise an InputError.
    """
    for key, value in data.items():
        if key not in DEFAULTS:
            raise InputError(f"Unknown configuration key: {key}")
        # The linter suggests to use "isinstance" here but this fails to
        # detect the difference between `int` and `bool`.
        if type(value) is not type(DEFAULTS[key]):
            expected_type = type(DEFAULTS[key]).__name__
            raise InputError(f"Data type for {key} must be {expected_type!r}")


def _check_output_config(config):
    """
    Run sanity checks on the generated config after all parsing and
    preprocessing is done.

    Raise InputError if an error is encountered.
    """
    if not config["paths"]:
        raise InputError("Please pass at least one file or directory")


def _parse_toml(infile):
    """
    Parse a TOML file for config values.

    It will search for a section named ``[tool.vulture]`` which contains the
    same keys as the CLI arguments seen with ``--help``. All leading dashes are
    removed and other dashes are replaced by underscores (so ``--sort-by-size``
    becomes ``sort_by_size``).

    Arguments containing multiple values are standard TOML lists.

    Example::

        [tool.vulture]
        exclude = ["file*.py", "dir/"]
        ignore_decorators = ["deco1", "deco2"]
        ignore_names = ["name1", "name2"]
        make_whitelist = true
        min_confidence = 10
        sort_by_size = true
        verbose = true
        paths = ["path1", "path2"]
    """
    data = tomllib.load(infile)
    settings = data.get("tool", {}).get("vulture", {})
    _check_input_config(settings)
    return settings


def _parse_args(args=None):
    """
    Parse CLI arguments.

    :param args: A list of strings representing the CLI arguments. If left to
        the default, this will default to ``sys.argv``.
    """

    # Sentinel value to distinguish between "False" and "no default given".
    missing = object()

    def csv(exclude):
        return exclude.split(",")

    usage = "%(prog)s [options] [PATH ...]"
    version = f"vulture {__version__}"
    glob_help = "Patterns may contain glob wildcards (*, ?, [abc], [!abc])."
    parser = argparse.ArgumentParser(prog="vulture", usage=usage)
    parser.add_argument(
        "paths",
        nargs="*",
        metavar="PATH",
        default=missing,
        help="Paths may be Python files or directories. For each directory"
        " Vulture analyzes all contained *.py files.",
    )
    parser.add_argument(
        "--exclude",
        metavar="PATTERNS",
        type=csv,
        default=missing,
        help=f"Comma-separated list of path patterns to ignore (e.g.,"
        f' "*settings.py,docs,*/test_*.py,venv"). {glob_help} A PATTERN'
        f" without glob wildcards is treated as *PATTERN*. Patterns are"
        f" matched against absolute paths.",
    )
    parser.add_argument(
        "--ignore-decorators",
        metavar="PATTERNS",
        type=csv,
        default=missing,
        help=f"Comma-separated list of decorators. Functions and classes using"
        f' these decorators are ignored (e.g., "@app.route,@require_*").'
        f" {glob_help}",
    )
    parser.add_argument(
        "--ignore-names",
        metavar="PATTERNS",
        type=csv,
        default=missing,
        help=f'Comma-separated list of names to ignore (e.g., "visit_*,do_*").'
        f" {glob_help}",
    )
    parser.add_argument(
        "--make-whitelist",
        action="store_true",
        default=missing,
        help="Report unused code in a format that can be added to a"
        " whitelist module.",
    )
    parser.add_argument(
        "--min-confidence",
        type=int,
        default=missing,
        help="Minimum confidence (between 0 and 100) for code to be"
        " reported as unused.",
    )
    parser.add_argument(
        "--sort-by-size",
        action="store_true",
        default=missing,
        help="Sort unused functions and classes by their lines of code.",
    )
    parser.add_argument(
        "--config",
        type=str,
        default="pyproject.toml",
        help="Path to pyproject.toml config file.",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", default=missing
    )
    parser.add_argument("--version", action="version", version=version)
    namespace = parser.parse_args(args)
    cli_args = {
        key: value
        for key, value in vars(namespace).items()
        if value is not missing
    }
    _check_input_config(cli_args)
    return cli_args


def make_config(argv=None, tomlfile=None):
    """
    Returns a config object for vulture, merging both ``pyproject.toml`` and
    CLI arguments (CLI arguments have precedence).

    :param argv: The CLI arguments to be parsed. This value is transparently
        passed through to :py:meth:`argparse.ArgumentParser.parse_args`.
    :param tomlfile: An IO instance containing TOML data. By default, this will
        auto-detect an existing ``pyproject.toml`` file and exists solely for
        unit-testing.
    """

    # Parse CLI first to skip sanity checks when --version or --help is given.
    cli_config = _parse_args(argv)

    # If we loaded data from a TOML file, we want to print this out on stdout
    # in verbose mode so we need to keep the value around.
    detected_toml_path = ""

    if tomlfile:
        config = _parse_toml(tomlfile)
        detected_toml_path = str(tomlfile)
    else:
        toml_path = pathlib.Path(cli_config["config"]).resolve()
        if toml_path.is_file():
            with open(toml_path, "rb") as fconfig:
                config = _parse_toml(fconfig)
            detected_toml_path = str(toml_path)
        else:
            config = {}

    # Overwrite TOML options with CLI options, if given.
    config.update(cli_config)

    # Set defaults for missing options.
    for key, value in DEFAULTS.items():
        config.setdefault(key, value)

    if detected_toml_path and config["verbose"]:
        print(f"Reading configuration from {detected_toml_path}")

    _check_output_config(config)

    return config



import ast
import pathlib
import sys
import tokenize
from enum import IntEnum


class VultureInputException(Exception):
    pass


class ExitCode(IntEnum):
    NoDeadCode = 0
    InvalidInput = 1
    InvalidCmdlineArguments = 2
    DeadCode = 3


def _safe_eval(node, default):
    """
    Safely evaluate the Boolean expression under the given AST node.

    Substitute `default` for all sub-expressions that cannot be
    evaluated (because variables or functions are undefined).

    We could use eval() to evaluate more sub-expressions. However, this
    function is not safe for arbitrary Python code. Even after
    overwriting the "__builtins__" dictionary, the original dictionary
    can be restored
    (https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html).

    """
    if isinstance(node, ast.BoolOp):
        results = [_safe_eval(value, default) for value in node.values]
        if isinstance(node.op, ast.And):
            return all(results)
        else:
            return any(results)
    elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
        return not _safe_eval(node.operand, not default)
    else:
        try:
            return ast.literal_eval(node)
        except ValueError:
            return default


def condition_is_always_false(condition):
    return not _safe_eval(condition, True)


def condition_is_always_true(condition):
    return _safe_eval(condition, False)


def is_ast_string(node):
    return isinstance(node, ast.Constant) and isinstance(node.value, str)


def format_path(path):
    try:
        return path.relative_to(pathlib.Path.cwd())
    except ValueError:
        # Path is not below the current directory.
        return path


def get_decorator_name(decorator):
    if isinstance(decorator, ast.Call):
        decorator = decorator.func
    try:
        parts = []
        while isinstance(decorator, ast.Attribute):
            parts.append(decorator.attr)
            decorator = decorator.value
        parts.append(decorator.id)
    except AttributeError:
        parts = []
    return "@" + ".".join(reversed(parts))


def get_modules(paths):
    """Retrieve Python files to check.

    Loop over all given paths, abort if any ends with .pyc, add the other given
    files (even those not ending with .py) and collect all .py files under the
    given directories.

    """
    modules = []
    for path in paths:
        path = path.resolve()
        if path.is_file():
            if path.suffix == ".pyc":
                sys.exit(f"Error: *.pyc files are not supported: {path}")
            else:
                modules.append(path)
        elif path.is_dir():
            modules.extend(path.rglob("*.py"))
        else:
            sys.exit(f"Error: {path} could not be found.")
    return modules


def read_file(filename):
    try:
        # Use encoding detected by tokenize.detect_encoding().
        with tokenize.open(filename) as f:
            return f.read()
    except (SyntaxError, UnicodeDecodeError) as err:
        raise VultureInputException from err


class LoggingList(list):
    def __init__(self, typ, verbose):
        super().__init__()
        self.typ = typ
        self._verbose = verbose

    def append(self, item):
        if self._verbose:
            print(f'define {self.typ} "{item.name}"')
        super().append(item)


class LoggingSet(set):
    def __init__(self, typ, verbose):
        super().__init__()
        self.typ = typ
        self._verbose = verbose

    def add(self, name):
        if self._verbose:
            print(f'use {self.typ} "{name}"')
        super().add(name)



import ssl

ctx = ssl.SSLContext()
ctx.check_hostname
ctx.hostname_checks_common_name
ctx.keylog_filename
ctx.maximum_version
ctx.minimum_version
ctx.options
ctx.post_handshake_auth
ctx.sni_callback
ctx.sslobject_class
ctx.sslsocket_class
ctx.verify_flags
ctx.verify_mode



import ast

from vulture import utils


class Reachability:
    def __init__(self, report):
        self._report = report
        self._no_fall_through_nodes = set()

        # Since we visit the children nodes first, we need to maintain a flag
        # that indicates if a break statement was seen. When visiting the
        # parent (While, For or AsyncFor), the value is checked (for While)
        # and reset. Assumes code is valid (break statements only in loops).
        self._current_loop_has_break_statement = False

    def visit(self, node):
        """When called, all children of this node have already been visited."""
        if isinstance(node, (ast.Break, ast.Continue, ast.Return, ast.Raise)):
            self._mark_as_no_fall_through(node)
            if isinstance(node, ast.Break):
                self._current_loop_has_break_statement = True

        elif isinstance(
            node,
            (
                ast.Module,
                ast.FunctionDef,
                ast.AsyncFunctionDef,
                ast.With,
                ast.AsyncWith,
            ),
        ):
            self._can_fall_through_statements_analysis(node.body)
        elif isinstance(node, ast.While):
            self._handle_reachability_while(node)
            self._current_loop_has_break_statement = False
        elif isinstance(node, (ast.For, ast.AsyncFor)):
            self._can_fall_through_statements_analysis(node.body)
            self._current_loop_has_break_statement = False
        elif isinstance(node, ast.If):
            self._handle_reachability_if(node)
        elif isinstance(node, ast.IfExp):
            self._handle_reachability_if_expr(node)
        elif isinstance(node, ast.Try):
            self._handle_reachability_try(node)

    def reset(self):
        self._no_fall_through_nodes = set()

    def _can_fall_through(self, node):
        return node not in self._no_fall_through_nodes

    def _mark_as_no_fall_through(self, node):
        self._no_fall_through_nodes.add(node)

    def _can_fall_through_statements_analysis(self, statements):
        """Report unreachable statements.
        Return True if we can execute the full list of statements.
        """
        for idx, statement in enumerate(statements):
            if not self._can_fall_through(statement):
                try:
                    next_sibling = statements[idx + 1]
                except IndexError:
                    next_sibling = None
                if next_sibling is not None:
                    class_name = statement.__class__.__name__.lower()
                    self._report(
                        name=class_name,
                        first_node=next_sibling,
                        last_node=statements[-1],
                        message=f"unreachable code after '{class_name}'",
                    )
                return False
        return True

    def _handle_reachability_if(self, node):
        has_else = bool(node.orelse)

        if utils.condition_is_always_false(node.test):
            self._report(
                name="if",
                first_node=node,
                last_node=node.body
                if isinstance(node, ast.IfExp)
                else node.body[-1],
                message="unsatisfiable 'if' condition",
            )
            if_can_fall_through = True
            else_can_fall_through = self._can_else_fall_through(
                node.orelse, condition_always_true=False
            )

        elif utils.condition_is_always_true(node.test):
            if_can_fall_through = self._can_fall_through_statements_analysis(
                node.body
            )
            else_can_fall_through = self._can_else_fall_through(
                node.orelse, condition_always_true=True
            )

            if has_else:
                self._report(
                    name="else",
                    first_node=node.orelse[0],
                    last_node=node.orelse[-1],
                    message="unreachable 'else' block",
                )
            else:
                # Redundant if-condition without else block.
                self._report(
                    name="if",
                    first_node=node,
                    message="redundant if-condition",
                )
        else:
            if_can_fall_through = self._can_fall_through_statements_analysis(
                node.body
            )
            else_can_fall_through = self._can_else_fall_through(
                node.orelse, condition_always_true=False
            )

        statement_can_fall_through = (
            if_can_fall_through or else_can_fall_through
        )

        if not statement_can_fall_through:
            self._mark_as_no_fall_through(node)

    def _can_else_fall_through(self, or_else, condition_always_true):
        if not or_else:
            return not condition_always_true
        return self._can_fall_through_statements_analysis(or_else)

    def _handle_reachability_if_expr(self, node):
        if utils.condition_is_always_false(node.test):
            self._report(
                name="ternary",
                first_node=node,
                last_node=node.body
                if isinstance(node, ast.IfExp)
                else node.body[-1],
                message="unsatisfiable 'ternary' condition",
            )
        elif utils.condition_is_always_true(node.test):
            else_body = node.orelse
            self._report(
                name="ternary",
                first_node=else_body,
                message="unreachable 'else' expression",
            )

    def _handle_reachability_while(self, node):
        if utils.condition_is_always_false(node.test):
            self._report(
                name="while",
                first_node=node,
                last_node=node.body
                if isinstance(node, ast.IfExp)
                else node.body[-1],
                message="unsatisfiable 'while' condition",
            )

        elif utils.condition_is_always_true(node.test):
            else_body = node.orelse
            if else_body:
                self._report(
                    name="else",
                    first_node=else_body[0],
                    last_node=else_body[-1],
                    message="unreachable 'else' block",
                )

            if not self._current_loop_has_break_statement:
                self._mark_as_no_fall_through(node)

        self._can_fall_through_statements_analysis(node.body)

    def _handle_reachability_try(self, node):
        try_can_fall_through = self._can_fall_through_statements_analysis(
            node.body
        )

        has_else = bool(node.orelse)

        if not try_can_fall_through and has_else:
            else_body = node.orelse
            self._report(
                name="else",
                first_node=else_body[0],
                last_node=else_body[-1],
                message="unreachable 'else' block",
            )

        any_except_can_fall_through = any(
            self._can_fall_through_statements_analysis(handler.body)
            for handler in node.handlers
        )

        statement_can_fall_through = (
            try_can_fall_through or any_except_can_fall_through
        )

        if not statement_can_fall_through:
            self._mark_as_no_fall_through(node)



#! /usr/bin/env python3

import re
import sys

_, VERSION, CHANGELOG, LIST = sys.argv
HEADER_REGEX = rf"# {VERSION} \(\d\d\d\d-\d\d-\d\d\)\n"

notes_list = []


with open(CHANGELOG) as f:
    first_line = next(f)
    if not re.match(HEADER_REGEX, first_line):
        sys.exit(
            f'First changelog line "{first_line.rstrip()}" must '
            f'start with "{HEADER_REGEX.rstrip()}"'
        )
    notes_list.extend([first_line[2:], "\n"])
    line = next(f)  # Skip empty line.
    if line.strip():
        sys.exit("There must be an empty line after each header")
    for line in f:
        if not line.strip():
            break
        else:
            notes_list.append(line)


def check(name, text):
    print("*" * 60)
    print(text)
    print("*" * 60)
    response = input(f"Accept this {name} (Y/n)? ").strip().lower()
    if response and response != "y":
        sys.exit(1)


check("changelog", "".join(notes_list))

with open(LIST, "w") as f:
    f.writelines(notes_list)



def get_last_line_number(node):
    return node.end_lineno


def get_first_line_number(node):
    """
    From Python 3.8 onwards, lineno for decorated objects is the line at which
    the object definition starts, which is different from what Python < 3.8
    reported -- the lineno of the first decorator. To preserve this behaviour
    of Vulture for newer Python versions, which is also more accurate for
    counting the size of the unused code chunk (if the property is unused, we
    also don't need it's decorators), we return the lineno of the first
    decorator, if there are any.
    """
    decorators = getattr(node, "decorator_list", [])
    if decorators:
        return decorators[0].lineno
    return node.lineno



import re
from collections import defaultdict

NOQA_REGEXP = re.compile(
    # Use the same regex as flake8 does.
    # https://github.com/pycqa/flake8/blob/main/src/flake8/defaults.py
    # We're looking for items that look like this:
    # `# noqa`
    # `# noqa: E123`
    # `# noqa: E123,W451,F921`
    # `# NoQA: E123,W451,F921`
    r"# noqa(?::\s?(?P<codes>([A-Z]+[0-9]+(?:[,\s]+)?)+))?",
    re.IGNORECASE,
)

NOQA_CODE_MAP = {
    # flake8 F401: module imported but unused.
    "F401": "V104",
    # flake8 F841: local variable is assigned to but never used.
    "F841": "V107",
}


def _parse_error_codes(match):
    # If no error code is specified, add the line to the "all" category.
    return [
        c.strip() for c in (match.groupdict()["codes"] or "all").split(",")
    ]


def parse_noqa(code):
    noqa_lines = defaultdict(set)
    for lineno, line in enumerate(code, start=1):
        match = NOQA_REGEXP.search(line)
        if match:
            for error_code in _parse_error_codes(match):
                error_code = NOQA_CODE_MAP.get(error_code, error_code)
                noqa_lines[error_code].add(lineno)
    return noqa_lines


def ignore_line(noqa_lines, lineno, error_code):
    """Check if the reported line is annotated with "# noqa"."""
    return lineno in noqa_lines[error_code] or lineno in noqa_lines["all"]



from unittest import TestCase, mock

TestCase.setUp
TestCase.tearDown
TestCase.setUpClass
TestCase.tearDownClass
TestCase.run
TestCase.skipTest
TestCase.debug
TestCase.failureException
TestCase.longMessage
TestCase.maxDiff
TestCase.subTest

mock.Mock.return_value
mock.Mock.side_effect

mock.MagicMock.return_value
mock.MagicMock.side_effect
